=begin

This file is part of Teatime.

Teatime is free software: you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation,
version 3 of the License.

Teatime is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Teatime. If not, see http://www.gnu.org/licenses/.


Copyright 1998-2009 Detlef Reichl  detlef!reichl()gmx!org

=end


# uncomment for debugging
#$log = File.new '/home/ich/Desktop/applet.log', File::CREAT|File::TRUNC|File::RDWR, 0644

module Teatime
    CUP_NAMES = %w[cup_classic cup_glass  mug_soeren]
    class Applet
        include GetText
        GetText.bindtextdomain 'teatime'

        def initialize applet, run_in_window
            @applet = applet
            @run_in_window = run_in_window

            init_gst

            @conf = Teatime::Preferences::Conf.new applet
            @area = Gtk::DrawingArea.new
            @size = nil
            @highlight = false
            @drawing = false
            @drawing_time = -1.0
            @last_cup = -1
            @last_size = -1

            @area.add_events Gdk::Event::Mask::BUTTON_PRESS_MASK |
                             Gdk::Event::Mask::ENTER_NOTIFY_MASK |
                             Gdk::Event::Mask::LEAVE_NOTIFY_MASK

            @area.signal_connect(:button_press_event) { |button, event| on_button_press event }
            @area.signal_connect(:enter_notify_event) { @highlight = true; redraw }
            @area.signal_connect(:leave_notify_event) { @highlight = false; redraw }
            @area.signal_connect(:expose_event)       { redraw }
            @area.signal_connect(:size_allocate)      { |area, allocation| on_size_allocate area, allocation }
            @area.signal_connect(:realize)            { on_realize }
            
            tea = @conf['last_tea']
            @area.tooltip_text = _("%s (%is)") % [_(@conf['teas'][tea]), @conf['times'][tea]]
          
            @popup = Teatime::Popup.new
            set_menu
        end
        attr_reader :area


        def init_gst
            Gst.init

            @gst_pipeline = Gst::ElementFactory.make 'playbin', 'playbin'
            return unless @gst_pipeline

            @gst_pipeline.set_property 'video-sink', nil
            @gst_pipeline.set_property 'vis-plugin', nil
            @gst_pipeline.bus.add_watch do |bus, message|
                if message.type == Gst::Message::ERROR or message.type == Gst::Message::EOS
                    @gst_pipeline.state = Gst::State::READY
                end
                true
            end
        end


        def set_menu
            xml = "<popup name=\"button3\">" +
                  "  <menuitem name=\"Item 1\" verb=\"about\" _label=\"" +
                  _('About') + "\"" +
                  "     pixtype=\"stock\" pixname=\"gtk-about\"/>" +
                  "  <menuitem name=\"Item 2\" verb=\"pref\" _label=\"" +
                  _('Preferences ...') + "\"" +
                  "     pixtype=\"stock\" pixname=\"gtk-properties\"/>" +
                  "  <separator/>"

            verbs = [['about', Proc.new{callback_about}],
                     ['pref',  Proc.new{callback_preferences}]]

            teas = @conf['teas']
            times = @conf['times']

            0.upto(teas.length() - 1) do |i|
                xml += "  <menuitem name=\"Item %i\" verb=\"%s\" _label=\"%s %is\"/>" %
                          [i, 't' + i.to_s, _(teas[i]), times[i]]
                verbs << ['t' + i.to_s, Proc.new{restart_drawing i}]
            end
            xml += "</popup>"

            @applet.set_menu xml, verbs
        end


        def callback_about
            if @about_dialog
                @about_dialog.present
                return
            end
            @about_dialog = Gtk::AboutDialog.new
            @about_dialog.authors = ['Detlef Reichl']
            @about_dialog.name = 'Teatime'
            @about_dialog.program_name = 'Teatime'
            @about_dialog.version = Teatime::VERSION
            @about_dialog.comments = _("A\ntea-drawing-time-observer.\n" +
                                       "The most useful GNOME program ;-)")
            @about_dialog.translator_credits = _("This crappy English is from\n" +
                                                 "Detlef Reichl <detlef ! reichl () gmx ! org>")
            @about_dialog.website = "http://det.cable.nu/teatime/index.rbx?r=#{Teatime::VERSION}"
            @about_dialog.website_label = _("Visit the Teatime website\nhttp://det.cable.nu/teatime")
            @about_dialog.logo = Gdk::Pixbuf.new PIXMAP_PATH + '/teatime.png'
            @about_dialog.set_icon PIXMAP_PATH + '/teatime.png'
            @about_dialog.show_all
            @about_dialog.signal_connect(:response) do
                @about_dialog.destroy
                @about_dialog = nil;
            end
        end


        def callback_preferences
            if @preferences_dialog
                @preferences_dialog.present
                return
            end
            @preferences_dialog = Teatime::Preferences::Dialog.new @conf
            @preferences_dialog.signal_connect(:cup_changed) do
                load_images
                redraw
            end
            @preferences_dialog.show
            @preferences_dialog.signal_connect(:response) do
                @conf['times'] = @preferences_dialog.times
                set_menu
                @preferences_dialog.destroy
                @preferences_dialog = nil
            end

        end



        def on_button_press event
            return false unless event.button == 1
            restart_drawing
            true
        end


        def on_size_allocate area, allocation
            if @applet.orient == PanelApplet::ORIENT_UP or @applet.orient == PanelApplet::ORIENT_DOWN
                @size = allocation.height
            else
                @size = allocation.width
            end
            if @size
                area.allocation.width = allocation.width
                area.allocation.height = allocation.height
                @area.set_size_request @size, @size
                load_images
            end
            if @backbuffer
                create_backbuffer
            end
        end


        def on_realize
            @backbuffer_gc = Gdk::GC.new @area.window
            create_backbuffer
        end

        def create_backbuffer
            @backbuffer = Gdk::Pixmap.new @area.window, @size, @size, -1
            @cairo = @backbuffer.create_cairo_context
            @cairo.set_source_rgba 1.0, 0.0, 0.0, 0.6
            @cairo.set_line_width 2.0
        end


        def restart_drawing tea_index=@conf['last_tea']
            if @drawing
                dialog = Gtk::MessageDialog.new nil,
                                                Gtk::Dialog::DESTROY_WITH_PARENT | Gtk::Dialog::MODAL,
                                                Gtk::MessageDialog::QUESTION,
                                                Gtk::MessageDialog::BUTTONS_YES_NO,
                                                _("Restart drawing?")
                dialog.set_icon PIXMAP_PATH + '/teatime.png'
                ret = dialog.run
                dialog.destroy
                return unless ret == Gtk::Dialog::RESPONSE_YES
                Gtk.timeout_remove @drawing_timeout
            end
            @conf['last_tea'] = tea_index
            @drawing = true
            @image_drawing = 1
            time = @conf['times'][tea_index]
            tea = @conf['teas'][tea_index]
            @drawing_time = time
            last_time = nil

            @drawing_timeout = Gtk.timeout_add(333.333) do
                @drawing_time -= 0.333334
                if @drawing_time <= 0.0
                    @area.tooltip_text = _("%s (%is)") % [_(tea), time]
                    Gtk.timeout_remove @drawing_timeout
                    @drawing = false
                    @image_drawing = 0
                    @popup.show if @conf['popup']
                    play_sound
                else
                    t = @drawing_time.to_i
                    if t != last_time
                        @area.tooltip_text = _("%s %i s left") %  [_(tea), t]
                        last_time = t
                    end
                    @image_drawing += 1
                    @image_drawing = 1 if @image_drawing == 4
                end
                redraw
                Gtk.main_iteration while Gtk.events_pending?
                true
            end
        end

        def play_sound
            return unless @gst_pipeline and @conf['play_sound']
            sound = @conf['sound']
            return if sound == ''

            state = @gst_pipeline.get_state
            if state == Gst::State::PLAYING or state == Gst::State::PAUSED
                @gst_pipeline.get_state Gst::ClockTime::NONE if @gst_pipeline.ready() == Gst::State::ASYNC
            end

            uri = GLib.filename_to_uri sound
            @gst_pipeline.set_property 'uri', uri
            @gst_pipeline.state = Gst::State::PLAYING
        end


        def redraw
            if @size
                image_num = @drawing ? @image_drawing : 0
                if @highlight
                    image = @images_applet_ligthen[image_num]
                else
                    image = @images_applet_normal[image_num]
                end

                @backbuffer.draw_pixbuf nil, image, 0, 0, 0, 0, -1, -1, Gdk::RGB::Dither::NONE, 0, 0
                if @drawing and @conf['show_clock']
                    ac = Math::PI * 1.5
                    arc = Math::PI * 2.0 * @drawing_time / @conf['times'][@conf['last_tea']]
                    hsize = @size / 2.0
                    @cairo.arc hsize, hsize, hsize - 1.0, ac, ac + arc
                    @cairo.stroke
                end
                @area.window.draw_drawable @backbuffer_gc, @backbuffer, 0, 0, 0, 0, -1, -1
            end
        end

        def load_images
            cup_index = @conf['cup']
            return if @last_cup == cup_index and @last_size == @size
            @last_cup = cup_index
            @last_size = @size

            @images_applet_normal = []
            @images_applet_ligthen = []
            path = PIXMAP_PATH + '/applet/'

            ['empty', 'full_0', 'full_1', 'full_2'].each do |i|
                cup = Teatime::CUP_NAMES[cup_index]
                pixbuf = Gdk::Pixbuf.new "#{path}#{cup}_#{i}.png"
                @images_applet_normal << pixbuf.scale(@size, @size, Gdk::Pixbuf::INTERP_HYPER)
                @images_applet_ligthen << lighten(pixbuf.scale(@size, @size, Gdk::Pixbuf::INTERP_HYPER))
            end
        end


        def lighten pixbuf
            return pixbuf if pixbuf.bits_per_sample != 8
            src = pixbuf.pixels
            dest = ''

            c4 = pixbuf.n_channels == 4
            rowdiff = pixbuf.rowstride - @size * (c4 ? 4 : 3)

            i = 0
            0.upto(@size -1) do
                0.upto(@size -1) do
                    dest << [src[i+0] + 32, 255].min
                    dest << [src[i+1] + 32, 255].min
                    dest << [src[i+2] + 32, 255].min
                    i += ((i + 3) % 4 != 0 and c4) ? 4 : 3
                end
                i += rowdiff
            end
            Gdk::Pixbuf.new dest, Gdk::Pixbuf::ColorSpace::RGB, false, 8, @size, @size, @size * 3
        end
    end
    

    class Popup
        def initialize
            @popup_image = 0
            load_images
            @window = Gtk::Window.new Gtk::Window::POPUP
            @window.set_default_size(*@masks[0].size)
            @window.resizable = false
            @window.window_position = Gtk::Window::POS_CENTER_ALWAYS

            eventbox = Gtk::EventBox.new
            @window << eventbox

            hbox = Gtk::HBox.new
            @images.each { |image| hbox.pack_start image, false, false, 0 }

            eventbox << hbox
            eventbox.signal_connect(:button_release_event) { hide }
            hbox.show
            eventbox.show
        end
        attr_reader :window


        def load_images
            @images = []
            @masks = []
            path = PIXMAP_PATH + '/popup/'
            0.upto(23) do |i|
                pixbuf = Gdk::Pixbuf.new path + sprintf("%02i.png", i)
                pixmap, mask = *pixbuf.render_pixmap_and_mask(128)
                @masks << mask
                @images << Gtk::Image.new(pixmap, mask)
            end
        end
        

        def show
            @window.present
            @timeout = Gtk.timeout_add(60) do
                @images[@popup_image].hide
                @popup_image += 1
                @popup_image = 0 if @popup_image >= 24
                @window.shape_combine_mask @masks[@popup_image], 0, 0
                @images[@popup_image].show
                true
            end
        end


        def hide
            @window.hide
            Gtk.timeout_remove @timeout
        end
    end
end


